
#define NB_MAX_PICS 128

#include <iostream.h>
#include <fstream.h>

extern "C"
{
#include "math.h"
#include "string.h"
}

#include "pictures.h"

int main(int argc, char **argv)
{
  Picture *pictures[NB_MAX_PICS], *result, *ref_pic;
  char result_name[64], ref_name[64];
  int nb_pics, x, y, d;
  float av, av0, te, te0, sum, sum_sq, v;
  float *array;
  int k, width, height;

  strcpy(result_name, "./result.ppm");
  strcpy(ref_name, "");
  nb_pics = 0;

  while(argc > 1)
    {
      if(strcmp(argv[1], "-p")==0)
	{
	  argv += 1; argc -= 1;
	  while(argc > 1)
	    {
	      if(nb_pics == NB_MAX_PICS)
		 {
		   cerr<<"Too many pics.\n";
		   exit(1);
		 }
	      pictures[nb_pics] = new Picture;
	      cerr<<"Loading "<<argv[1]<<" ... "; cerr.flush();
	      pictures[nb_pics]->LoadPpm(argv[1]);
	      cerr<<"done.\n";
	      if(nb_pics==0)
		{
		  width = pictures[nb_pics]->SizeX;
		  height = pictures[nb_pics]->SizeY;
		}
	      else
		{
		  if((width != pictures[nb_pics]->SizeX) ||
		     (height != pictures[nb_pics]->SizeY))
		    {
		      cerr<<"Size mismatch.\n";
		      exit(1);
		    }
		}
	      nb_pics++;
	      argv += 1; argc -= 1;
	    }
	}
      else if(strcmp(argv[1], "-o")==0)
	{
	  if(argc>2)
	    {
	      strncpy(result_name, argv[2], 64);
	      argv += 2; argc -= 2;
	      cout<<"Output sent to "<<result_name<<".\n";
	    }
	  else
	    {
	      cerr<<"No filename.\n";
	      exit(1);
	    }
	}
      else if(strcmp(argv[1], "-r")==0)
	{
	  if(argc>2)
	    {
	      strncpy(ref_name, argv[2], 64);
	      argv += 2; argc -= 2;
	      cout<<"Using "<<ref_name<<" as reference picture.\n";
	    }
	  else
	    {
	      cerr<<"No filename.\n";
	      exit(1);
	    }
	}
      else
	{
	  cerr<<"Unknown option "<<argv[1]<<".\n";
	  exit(1);
	}
    }

  cerr<<"Width="<<width<<" Height="<<height<<"\n";
  cerr<<"Smoothing ... "; cerr.flush();

  array = new float[nb_pics];

  result = new Picture(width, height);

  for(y=0; y<height; y++) for(x=0; x<width; x++)
    {
      sum = 0;
      for(k=0; k<nb_pics; k++)
	sum += pictures[k]->GetComponent(x, y, PIC_RED);
      for(d=0; d<PIC_DEPTH; d++)
	result->SetComponent(x, y, d, sum/float(nb_pics));
    }

  cerr<<"done.\n";
  
  if(strlen(ref_name) > 0)
    {
      cerr<<"Lightning correction ... "; cerr.flush();
      ref_pic = new Picture;
      ref_pic->LoadPpm(ref_name);

      sum = 0; sum_sq = 0;
      for(x=0; x<width; x++) for(y=0; y<height; y++)
	{
	  v = ref_pic->GetComponent(x, y, PIC_RED);
	  sum += v;
	  sum_sq += v*v;
	}
      av0 = sum/float(width * height);
      te0 = sqrt(sum_sq/float(width * height) - av0*av0);

      for(x=0; x<width; x++)
	{
	  sum = 0; sum_sq = 0;
	  for(y=0; y<height; y++)
	    {
	      v = ref_pic->GetComponent(x, y, PIC_RED);
	      sum += v;
	      sum_sq += v*v;
	    }
	  
	  av = sum/float(height);
	  te = sqrt(sum_sq/float(height) - av*av);
	  
	  for(y=0; y<height; y++)
	    {
	      v = result->GetComponent(x, y, PIC_RED);
	      //if(te>0) v = av0 + (v - av)*(te0/te);
	      //else 
	      v = av0 + (v - av);
	      if(v < 0) v = 0; else if(v > 1) v = 1;
	      for(d=0; d<PIC_DEPTH; d++) result->SetComponent(x, y, d, v);
	    }
	}
      delete ref_pic;
      cerr<<"done.\n";
    }
  else cerr<<"No column-lightning correction.\n";

  for(k=0; k<nb_pics; k++) delete pictures[k];

  cerr<<"Saving ... "; cerr.flush();
  result->SavePpm(result_name);
  cerr<<"done.\n";
  delete result;
}
